<?php

/**
 * @package     EasyStore.Site
 * @subpackage  com_easystore
 *
 * @copyright   (C) 2023 JoomShaper. <https://www.joomshaper.com>
 * @license     GNU General Public License version 2 or later; see LICENSE.txt
 */

namespace JoomShaper\Component\EasyStore\Site\Controller;

use Exception;
use Throwable;
use Joomla\CMS\Factory;
use Joomla\CMS\Uri\Uri;
use Joomla\Input\Input;
use Joomla\CMS\Router\Route;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Event\Table\AbstractEvent;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\MVC\Controller\BaseController;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use JoomShaper\Component\EasyStore\Site\Traits\Api;
use JoomShaper\Component\EasyStore\Site\Email\Email;
use JoomShaper\Component\EasyStore\Site\Model\CartModel;
use JoomShaper\Component\EasyStore\Site\Traits\Checkout;
use JoomShaper\Component\EasyStore\Site\Helper\ArrayHelper;
use JoomShaper\Component\EasyStore\Site\Model\CheckoutModel;
use JoomShaper\Component\EasyStore\Site\Model\SettingsModel;
use JoomShaper\Component\EasyStore\Site\Helper\EasyStoreHelper;
use JoomShaper\Component\EasyStore\Administrator\Helper\SettingsHelper;
use JoomShaper\Component\EasyStore\Administrator\Helper\EasyStoreDatabaseOrm;
use JoomShaper\Component\EasyStore\Administrator\Model\OrderModel as AdminOrderModel;

// phpcs:disable PSR1.Files.SideEffects
\defined('_JEXEC') or die;
// phpcs:enable PSR1.Files.SideEffects

/**
 * Checkout Controller of EasyStore component
 *
 * @since  1.0.0
 */
class CheckoutController extends BaseController
{
    use Api;
    use Checkout;

    public function __construct($config = [], MVCFactoryInterface $factory = null, ?CMSApplication $app = null, ?Input $input = null)
    {
        parent::__construct($config, $factory, $app, $input);
    }

    protected function getDefaultCountryState()
    {
        $customer = EasyStoreHelper::getCustomerByUserId(Factory::getApplication()->getIdentity()->id);

        $defaultCountry = null;
        $defaultState   = null;

        if (!empty($customer->shipping_address)) {
            $shippingAddress = is_string($customer->shipping_address) ? json_decode($customer->shipping_address) : null;

            if ($shippingAddress) {
                $defaultCountry = $shippingAddress->country ?? null;
                $defaultState   = $shippingAddress->state ?? null;
            }
        }

        return [$defaultCountry, $defaultState];
    }

    public function getCartData()
    {
        [$defaultCountry, $defaultState] = $this->getDefaultCountryState();
        $cartModel                       = new CartModel();
        $shippingId                      = Factory::getApplication()->input->get('shipping_id', null, 'STRING');
        $country                         = Factory::getApplication()->input->get('country', $defaultCountry, 'STRING');
        $state                           = Factory::getApplication()->input->get('state', $defaultState, 'STRING');
        $cart                            = $cartModel->getItem($shippingId, $country, $state);

        $this->sendResponse($cart);
    }

    public function getInformation()
    {
        $model       = new CheckoutModel();
        $information = $model->getInformation();

        $this->sendResponse($information);
    }

    public function getCheckoutSettings()
    {
        $settings        = SettingsHelper::getSettings();
        $checkoutSetting = $settings->get('checkout', null) ?? (object) [];

        $this->sendResponse($checkoutSetting);
    }


    public function getShipping()
    {
        [$defaultCountry, $defaultState] = $this->getDefaultCountryState();

        $input    = Factory::getApplication()->input;
        $country  = $input->get('country', $defaultCountry);
        $state    = $input->get('state', $defaultState);
        $subtotal = $input->get('subtotal', null);

        $settingModel = new SettingsModel();

        $this->sendResponse($settingModel->getShipping($country, $state, $subtotal) ?? []);
    }

    public function getPaymentMethods()
    {
        $this->sendResponse($this->getActivePayments());
    }

    public function saveShipping($data)
    {
        $orm = new EasyStoreDatabaseOrm();

        try {
            $orm->update('#__easystore_cart', $data, 'token');

            return true;
        } catch (Throwable $error) {
            throw $error;
        }
    }

    public function searchGuestUser()
    {
        $input = Factory::getApplication()->input;
        $email = $input->get('email', '', 'email');

        $model    = new CheckoutModel();
        $shipping = $model->getGuestShippingAddress($email);

        $this->sendResponse($shipping);
    }

    public function placeOrder()
    {
        $user                 = Factory::getApplication()->getIdentity();
        $input                = Factory::getApplication()->input;
        $token                = (new CartController())->getToken();
        $settings             = SettingsHelper::getSettings();
        $allowGuestCheckout   = $settings->get('checkout.allow_guest_checkout', false);
        $isGuestCheckout      = $user->guest && $allowGuestCheckout;
        $saveGuestInformation = (bool) $input->get('save_address', false);
        $customerEmail        = $input->get('email', '', 'email');
        $customerEmail        = $isGuestCheckout ? $customerEmail : $user->email;

        $cartItems = (new CartModel())->getItem();

        $data = (object) [
            'token'          => $token,
            'payment_method' => $input->get('payment_method', '', 'RAW'),
            'cart_items'     => $cartItems,
        ];

        $shippingAddress = $input->get('shipping_address', '', 'RAW') ?? "[]";

        if (!empty($shippingAddress)) {
            $shippingAddress = json_decode($shippingAddress);

            if (isset($shippingAddress->country) && empty($shippingAddress->country)) {
                $this->sendResponse(['message' => Text::_('COM_EASYSTORE_CART_SHIPPING_COUNTRY_NOT_SELECTED')], 500);

                return false;
            }
        }

        if (!$input->get('is_billing_and_shipping_address_same', 0, 'INT')) {
            $billingAddress = $input->get('billing_address', '', 'RAW') ?? "[]";

            if (!empty($billingAddress)) {
                $billingAddress = json_decode($billingAddress);

                if (isset($billingAddress->country) && empty($billingAddress->country)) {
                    $this->sendResponse(['message' => Text::_('COM_EASYSTORE_CART_BILLING_COUNTRY_NOT_SELECTED')], 500);

                    return false;
                }
            }
        }

        $information = (object) [
            'user_id'                              => $user->id ?? 0,
            'shipping_address'                     => $input->get('shipping_address', '', 'RAW'),
            'is_billing_and_shipping_address_same' => $input->get('is_billing_and_shipping_address_same', 0, 'INT'),
            'billing_address'                      => $input->get('billing_address', '', 'RAW'),
        ];

        $phoneNumber = $input->get('phone', '', 'STRING');
        if (!empty($phoneNumber)) {
            $information->phone = $phoneNumber;
        }

        $shippingData = (object) [
            'token'           => $token,
            'shipping_method' => $input->get('shipping_method', '', 'RAW'),
        ];

        if ($information->is_billing_and_shipping_address_same) {
            $information->billing_address = $information->shipping_address;
        }

        $model = new CheckoutModel();

        $orm = new EasyStoreDatabaseOrm();
        $db  = Factory::getDbo();

        $db->transactionStart();

        try {
            if ($isGuestCheckout) {
                if ($saveGuestInformation) {
                    $guestData = (object) [
                        'email'            => $customerEmail,
                        'shipping_address' => $information->shipping_address,
                    ];

                    $model->saveGuestCustomerInformation($guestData);
                }
            } else {
                $model->saveCustomerInformation($information);
            }

            $this->saveShipping($shippingData);

            $orm->update('#__easystore_cart', $data, 'token');
            $redirectionUrl = null;

            $orderData = (object) [
                'order_status'      => 'active',
                'email'             => $customerEmail,
                'shipping_address'  => $information->shipping_address,
                'billing_address'   => $information->billing_address,
                'is_guest_checkout' => $isGuestCheckout,
            ];

            $order = $this->createOrder($orderData);

            $paymentMethodsWithoutCheckoutUrl = $this->getPaymentMethodsWithoutCheckoutUrl();

            if ($data->payment_method !== 'manual_payment' && !in_array($data->payment_method, $paymentMethodsWithoutCheckoutUrl)) {
                $redirectionUrl = $this->navigateToPaymentGateway($data, $shippingAddress, $order->id);
            } elseif (in_array($data->payment_method, $paymentMethodsWithoutCheckoutUrl)) {
                $json_base64 = base64_encode(json_encode(['orderID' => $order->id, 'shippingAddress' => $order->shipping_address]));

                $redirectionUrl = Route::_('index.php?option=com_easystore&task=payment.onPayment&type=' . $data->payment_method . '&data=' . $json_base64, false);
            } else {
                $redirectionUrl = Route::_('index.php?option=com_easystore&task=payment.onPaymentSuccess&type=' . $data->payment_method, false);
            }

            $db->transactionCommit();

            $subject    = Text::sprintf('COM_EASYSTORE_EMAIL_ORDER_CONFIRM', $order->id);

            $displayData = [
                'order_id'       => $order->id,
                'order_date'     => $order->creation_date,
                'payment_status' => $order->payment_status,
                'payment_method' => $order->payment_method,
                'order_link'     => Uri::root() . 'index.php?option=com_easystore&view=order&id=' . $order->id,
                'customer_name'  => $user->name ?? Text::_('COM_EASYSTORE_ORDER_BY_CUSTOMER'),
            ];

            $email = new Email($customerEmail, $subject, 'emails.order.created', $displayData);
            $email->send();

            $this->sendResponse(['success' => true, 'redirectionUrl' => $redirectionUrl]);
        } catch (Exception $error) {
            $db->transactionRollback();
            $this->sendResponse(['message' => $error->getMessage()], 500);
        }
    }

    public function onOrderPlacementCompletion()
    {
        $this->deductQuantityFromInventory();
        $this->removeCartData();
    }

    public function deductQuantityFromInventory()
    {
        $cart = (new CartModel())->getItem();

        if (empty($cart)) {
            return;
        }

        if (empty($cart->items)) {
            return;
        }

        foreach ($cart->items as $item) {
            $data = (object) [
                'product_id' => $item->product_id,
                'sku_id'     => $item->sku_id,
                'quantity'   => $item->quantity,
            ];

            $this->deductFromProductOrSkuTable($data);
        }
    }

    public function deductFromProductOrSkuTable($data)
    {
        $db    = Factory::getDbo();
        $query = $db->getQuery(true);
        $query->select('is_tracking_inventory')->from($db->quoteName('#__easystore_products'))
            ->where($db->quoteName('id') . ' = ' . (int) $data->product_id);

        $db->setQuery($query);

        try {
            $isTrackingInventory = (bool) ($db->loadResult() ?? 0);
        } catch (Throwable $error) {
            throw $error;
        }


        if (!$isTrackingInventory) {
            return;
        }

        $tableName  = !empty($data->sku_id) ? '#__easystore_product_skus' : '#__easystore_products';
        $columnName = !empty($data->sku_id) ? 'inventory_amount' : 'quantity';

        $fields = [
            $db->quoteName($columnName) . ' = ' . $db->quoteName($columnName) . ' - ' . $data->quantity,
        ];

        $conditions = [
            $db->quoteName('id') . ' = ' . $db->quote(!empty($data->sku_id) ? $data->sku_id : $data->product_id),
        ];

        $query = $db->getQuery(true)->update($tableName)
            ->set($fields)
            ->where($conditions);

        $db->setQuery($query);

        try {
            $db->execute();
        } catch (Throwable $error) {
            throw $error;
        }
    }

    public function removeCartData()
    {
        $cartController = new CartController();
        $token          = $cartController->getToken();

        $db    = Factory::getDbo();
        $query = $db->getQuery(true);

        $query->delete($db->quoteName('#__easystore_cart'))
            ->where($db->quoteName('token') . ' = ' . $db->quote($token));

        $db->setQuery($query);

        try {
            $db->execute();
            $cartController->removeToken();
        } catch (Throwable $error) {
            throw $error;
        }
    }

    public function navigateToPaymentGateway($data, $shippingAddress, $orderId)
    {
        $paymentMethod = $data->payment_method ?: '';

        if ($paymentMethod) {
            if (!PluginHelper::isEnabled('easystore', $paymentMethod)) {
                $message = Text::sprintf('COM_EASYSTORE_CART_PLUGIN_IS_DISABLE', ucfirst($paymentMethod), 'error');
                throw new Exception($message);
            }

            PluginHelper::importPlugin('easystore', $paymentMethod);

            $cartDataForPayment = EasyStoreHelper::getCartDataForPayment($shippingAddress);

            $event = AbstractEvent::create(
                'onPayment',
                [
                    'subject'        => (object)$cartDataForPayment,
                    'order_id'       => $orderId,
                    'redirectionUrl' => '',
                ]
            );

            try {
                $eventResult = Factory::getApplication()->getDispatcher()->dispatch($event->getName(), $event);

                return $eventResult->getArgument('redirectionUrl') ?: null;
            } catch (Throwable $error) {
                Factory::getApplication()->enqueueMessage($error->getMessage(), 'error');
            }
        }
    }

    public function createOrder($data)
    {
        [$defaultCountry, $defaultState] = $this->getDefaultCountryState();
        $shippingAddress = Factory::getApplication()->input->get('shipping_address', '', 'RAW');
        
        if (!empty($shippingAddress) && is_string($shippingAddress))
        {
            $shippingAddress = json_decode($shippingAddress);
        }

        $country = $shippingAddress->country ?? $defaultCountry;
        $state = $shippingAddress->state ?? $defaultState;

        $user            = Factory::getApplication()->getIdentity();
        $currentDateTime = Factory::getDate('now')->toSql();
        $customer        = EasyStoreHelper::getCustomerByUserId($user->id);
        $cartModel       = new CartModel();
        $cartItem        = $cartModel->getItem(null, $country, $state);

        $settings            = SettingsHelper::getSettings();
        $isTaxEnabled        = $settings->get('checkout.enable_tax', true);
        $isCouponCodeEnabled = $settings->get('checkout.enable_coupon_code', true);

        $data = (object) [
            'id'               => null,
            'creation_date'    => $currentDateTime,
            'customer_id'      => $customer->id ?? null,
            'customer_email'   => $data->email,
            'shipping_address' => $data->shipping_address,
            'billing_address'  => $data->billing_address,
            'customer_note'    => '',
            'payment_status'   => 'unpaid',
            'fulfilment'       => 'unfulfilled',
            'order_status'     => $data->order_status,
            'is_guest_order'   => $data->is_guest_checkout ? 1 : 0,
            'discount_type'    => 'percent',
            'discount_value'   => 0,
            'discount_reason'  => '',
            'shipping'         => !is_string($cartItem->shipping_method) ? json_encode($cartItem->shipping_method) : '',
            'payment_method'   => $cartItem->payment_method,
            'created'          => $currentDateTime,
            'created_by'       => $user->id,
        ];

        if ($isTaxEnabled) {
            $data->sale_tax = $cartItem->taxable_amount ?? 0.00;
        }

        if ($isCouponCodeEnabled) {
            $data->coupon_code   = $cartItem->coupon_code ?? null;
            $data->coupon_type   = $cartItem->coupon_type ?? null;
            $data->coupon_amount = $cartItem->coupon_amount ?? 0.00;
        }

        $orm = new EasyStoreDatabaseOrm();

        try {
            $orm->create('#__easystore_orders', $data, 'id');
            $this->addOrderItems($cartItem->items, $data->id);

            $adminOrderModel = new AdminOrderModel();
            $adminOrderModel->addOrderActivity($data->id, 'order_created');

            return $data;
        } catch (Throwable $error) {
            throw $error;
        }
    }

    public function updateOrder($data)
    {
        $orm = new EasyStoreDatabaseOrm();

        try {
            $orm->update('#__easystore_orders', $data, 'id');

            return true;
        } catch (Throwable $error) {
            throw $error;
        }

        return false;
    }

    public function addOrderItems($items, $orderId)
    {
        if (empty($items)) {
            return;
        }

        $orm = new EasyStoreDatabaseOrm();

        foreach ($items as $item) {
            $data = (object) [
                'order_id'        => $orderId,
                'product_id'      => $item->product_id,
                'variant_id'      => $item->sku_id ?? null,
                'discount_type'   => $item->discount->type ?? 'percent',
                'discount_value'  => $item->discount->value ?? 0,
                'discount_reason' => '',
                'quantity'        => $item->quantity,
                'price'           => $item->item_price,
            ];

            try {
                $orm->create('#__easystore_order_product_map', $data);
            } catch (Throwable $error) {
                throw $error;
            }
        }

        return true;
    }

    public function applyCoupon()
    {
        $input  = Factory::getApplication()->input;
        $code   = $input->get('code', '', 'STRING');
        $cartId = $input->get('cart_id', 0, 'INT');

        if (!$code || !$cartId) {
            $this->sendResponse(['message' => 'Missing required information.'], 400);
        }

        $model = new CheckoutModel();

        try {
            $coupon = $model->getCouponByCode($code);
        } catch (Exception $error) {
            $this->sendResponse(['message' => $error->getMessage()], 500);
        }

        if (is_null($coupon)) {
            $this->sendResponse(['message' => Text::_('COM_EASYSTORE_CART_COUPON_INVALID')], 404);
        }

        if (!EasyStoreHelper::isCouponCodeValid($coupon->start_date, $coupon->end_date)) {
            $this->sendResponse(['message' => Text::_('COM_EASYSTORE_CART_COUPON_CODE_EXPIRED')], 400);
        }

        $couponData = (object) [
            'code'   => $coupon->code,
            'type'   => $coupon->discount_type,
            'amount' => $coupon->discount_value,
        ];

        try {
            $model->applyCouponCode($cartId, $couponData);
        } catch (Exception $error) {
            $this->sendResponse(['message' => $error->getMessage()], 500);
        }

        $this->sendResponse(['success' => true]);
    }

    public function removeCode()
    {
        $input  = Factory::getApplication()->input;
        $cartId = $input->get('cart_id', 0, 'INT');

        if (!$cartId) {
            $this->sendResponse(['message' => 'Missing required information.'], 400);
        }

        $model = new CheckoutModel();

        try {
            $model->removeCouponCode($cartId);
        } catch (Exception $error) {
            $this->sendResponse(['message' => $error->getMessage()], 500);
        }

        $this->sendResponse(['success' => true]);
    }

    private function loadCountriesData()
    {
        $jsonPath  = JPATH_ROOT . '/media/com_easystore/data/countries.json';
        $countries = [];

        if (file_exists($jsonPath)) {
            $countries = file_get_contents($jsonPath);

            if (!empty($countries) && is_string($countries)) {
                $countries = json_decode($countries);
            }
        }

        return $countries;
    }


    public function getCountries()
    {
        $countriesData = $this->loadCountriesData();

        $settingsModel     = new SettingsModel();
        $countryWithStates = $settingsModel->getCountriesWithStates();

        $codes = array_keys($countryWithStates);

        $countries = ArrayHelper::findByArray(function ($haystack, $item) {
            return ArrayHelper::find(function ($value) use ($item) {
                return $value->numeric_code == $item;
            }, $haystack);
        }, $countriesData, $codes);

        $countries = ArrayHelper::toOptions(function ($country) {
            return (object)[
                'label' => $country->name,
                'value' => $country->numeric_code,
            ];
        }, $countries);

        $this->sendResponse($countries);
    }

    protected function isAllStatesAllowed($countryCode)
    {
        $settings = SettingsHelper::getSettings();
        $shipping = $settings->get('shipping', []);

        $isAllowed = false;

        foreach ($shipping as $item) {
            if (!empty($item->regions)) {
                foreach ($item->regions as $region) {
                    if ($region->country == $countryCode && empty($region->states)) {
                        $isAllowed = true;
                        break;
                    }
                }
            }

            if ($isAllowed) {
                break;
            }
        }

        return $isAllowed;
    }

    public function getStates()
    {
        $input       = Factory::getApplication()->input;
        $countryCode = $input->get('country_code', '');

        $settingsModel     = new SettingsModel();
        $countryWithStates = $settingsModel->getCountriesWithStates();

        if (empty($countryCode) || empty($countryWithStates)) {
            $this->sendResponse([]);
        }

        $countriesData   = $this->loadCountriesData();
        $country         = ArrayHelper::find(function ($item) use ($countryCode) {
            return $item->numeric_code == $countryCode;
        }, $countriesData);

        if (is_null($country)) {
            $this->sendResponse([]);
        }

        $states = $country->states ?? [];

        if (empty($states)) {
            $this->sendResponse([]);
        }

        if ($this->isAllStatesAllowed($countryCode)) {
            $options = ArrayHelper::toOptions(function ($item) {
                return (object) [
                    'label' => $item->name,
                    'value' => $item->id,
                ];
            }, $states);

            $this->sendResponse($options);
        }

        $stateCodes = $countryWithStates[$countryCode] ?? [];

        if (empty($stateCodes)) {
            $this->sendResponse([]);
        }

        $states = ArrayHelper::findByArray(function ($haystack, $item) {
            return ArrayHelper::find(function ($value) use ($item) {
                return $value->id == $item;
            }, $haystack);
        }, $states, $stateCodes);


        $options = ArrayHelper::toOptions(function ($item) {
            return (object)[
                'label' => $item->name,
                'value' => $item->id,
            ];
        }, $states);

        $this->sendResponse($options);
    }
}
