<?php

/**
 * @package     EasyStore.Site
 * @subpackage  com_easystore
 *
 * @copyright   (C) 2023 JoomShaper. <https://www.joomshaper.com>
 * @license     GNU General Public License version 2 or later; see LICENSE.txt
 */

namespace JoomShaper\Component\EasyStore\Site\Controller;

use Exception;
use Throwable;
use Joomla\CMS\Factory;
use Joomla\CMS\Router\Route;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Event\AbstractEvent;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\MVC\Controller\BaseController;
use JoomShaper\Component\EasyStore\Site\Traits\Api;
use JoomShaper\Component\EasyStore\Site\Model\CartModel;
use JoomShaper\Component\EasyStore\Site\Helper\EasyStoreHelper;
use JoomShaper\Component\EasyStore\Site\Controller\CheckoutController;

// phpcs:disable PSR1.Files.SideEffects
\defined('_JEXEC') or die;
// phpcs:enable PSR1.Files.SideEffects

/**
 * Default Controller of EasyStore component
 *
 * @since  1.0.0
 */
class PaymentController extends BaseController
{
    use Api;

    /**
     * After a successful payment, the following functionalities will be triggered.
     *
     * @since 1.0.0
     */
    public function onPaymentSuccess()
    {
        (new CheckoutController())->removeCartData();

        /** @var CMSApplication */
        $app         = Factory::getApplication();
        $input       = $app->getInput();
        $paymentType = $input->get('type', '', 'STRING');
        $type        = $paymentType ? '&type=' . $paymentType : '';

        $app->redirect(Route::_('index.php?option=com_easystore&view=payment&layout=success' . $type, false));
    }

    /**
     * If the payment is canceled, the following functionalities will be triggered.
     * @since 1.0.0
     * @return void
     */
    public function onPaymentCancel()
    {
        /** @var CMSApplication */
        $app = Factory::getApplication();
        $app->redirect(Route::_('index.php?option=com_easystore&view=payment&layout=cancel', false));
    }

    /**
     * This function manages payment notifications. It identifies the payment type, imports the relevant
     * plugin, and triggers the 'onPaymentNotify' event. If any errors occur, a user-friendly message is
     * displayed.
     *
     * @since 1.0.0
     */
    public function onPaymentNotify()
    {
        $paymentType = $this->input->getString('type');
        PluginHelper::importPlugin('easystore', $paymentType);

        $event = AbstractEvent::create(
            'onPaymentNotify',
            [
                'subject' => new \stdClass(),
            ]
        );

        try {
            Factory::getApplication()->getDispatcher()->dispatch($event->getName(), $event);
        } catch (\Throwable $th) {
            Factory::getApplication()->enqueueMessage($th->getMessage());
        }
    }

    /**
     * Handle the creation of a webhook for a specific payment type.
     * @since 1.0.0
     */
    public function onPaymentCreateWebhook()
    {
        $paymentType = $this->input->getString('type');
        PluginHelper::importPlugin('easystore', $paymentType);

        $event = AbstractEvent::create(
            'onCreateWebhook',
            [
                'subject' => new \stdClass(),
            ]
        );

        try {
            $eventResult = Factory::getApplication()->getDispatcher()->dispatch($event->getName(), $event);
            $message     = $eventResult['result'][0] ? $eventResult['result'][0] : 'already_created';

            $this->sendResponse(['success' => true, 'message' => $message]);
        } catch (\Throwable $error) {
            $this->sendResponse(['success' => false, 'message' => $error->getMessage()]);
        }
    }

    /**
     * Handles the onPayment event for a specific payment method which payments doest not sent any redirection URL.
     *
     * @throws Exception If the specified payment plugin is not enabled.
     * @since 1.0.2
    */
    public function onPayment()
    {
        $paymentMethod = $this->input->getString('type');
        $checkoutData  = $this->input->getString('data');

        $paymentMethod = $paymentMethod ?: '';

        if ($paymentMethod) {
            if (!PluginHelper::isEnabled('easystore', $paymentMethod)) {
                $message = Text::sprintf('COM_EASYSTORE_CART_PLUGIN_IS_DISABLE', ucfirst($paymentMethod), 'error');
                throw new Exception($message);
            }

            PluginHelper::importPlugin('easystore', $paymentMethod);

            $checkoutInfo = json_decode(base64_decode($checkoutData));

            $cartDataForPayment = EasyStoreHelper::getCartDataForPayment($checkoutInfo->shippingAddress);

            $event = AbstractEvent::create(
                'onPayment',
                [
                    'subject'  => (object) $cartDataForPayment,
                    'order_id' => $checkoutInfo->orderID,
                ]
            );

            try {
                Factory::getApplication()->getDispatcher()->dispatch($event->getName(), $event);
            } catch (Throwable $error) {
                Factory::getApplication()->enqueueMessage($error->getMessage(), 'error');
            }
        }
    }
}
